#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2011       Nick Hall
# Copyright (C) 2011       Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Module that implements the gramplet bar fuctionality.
"""

# -------------------------------------------------------------------------
#
# Set up logging
#
# -------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".grampletbar")

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import time
import os
import configparser

# -------------------------------------------------------------------------
#
# GNOME modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import Gdk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.const import URL_MANUAL_PAGE, VERSION_DIR
from gramps.gen.config import config
from gramps.gen.constfunc import win
from gramps.gen.utils.configmanager import clean_up
from ..managedwindow import ManagedWindow
from ..display import display_help, display_url
from .grampletpane import (
    AVAILABLE_GRAMPLETS,
    GET_AVAILABLE_GRAMPLETS,
    GET_GRAMPLET_LIST,
    get_gramplet_opts,
    get_gramplet_options_by_name,
    make_requested_gramplet,
    GuiGramplet,
)
from .undoablebuffer import UndoableBuffer
from ..utils import is_right_click
from ..dialog import QuestionDialog

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
WIKI_HELP_PAGE = URL_MANUAL_PAGE + "_-_Gramplets"
MAIN_HELP_PAGE = URL_MANUAL_PAGE + "_-_Main_Window"
NL = "\n"


# -------------------------------------------------------------------------
#
# GrampletBar class
#
# -------------------------------------------------------------------------
class GrampletBar(Gtk.Notebook):
    """
    A class which defines the graphical representation of the GrampletBar.
    """

    def __init__(
        self,
        dbstate,
        uistate,
        pageview,
        configfile,
        defaults,
        orientation=Gtk.Orientation.VERTICAL,
    ):
        Gtk.Notebook.__init__(self)

        self.dbstate = dbstate
        self.uistate = uistate
        self.pageview = pageview
        self.configfile = os.path.join(VERSION_DIR, "%s.ini" % configfile)
        self.defaults = defaults
        self.orientation = orientation
        self.detached_gramplets = []
        self.empty = False
        self.close_buttons = []

        self.set_group_name("grampletbar")
        self.set_show_border(False)
        self.set_scrollable(True)

        book_button = Gtk.Button()
        # Arrow is too small unless in a box
        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        arrow = Gtk.Arrow(
            arrow_type=Gtk.ArrowType.DOWN, shadow_type=Gtk.ShadowType.NONE
        )
        arrow.show()
        box.add(arrow)
        box.show()
        book_button.add(box)
        book_button.set_relief(Gtk.ReliefStyle.NONE)
        book_button.connect("clicked", self.__button_clicked)
        book_button.set_property("tooltip-text", _("Gramplet Bar Menu"))
        book_button.show()
        self.set_action_widget(book_button, Gtk.PackType.END)

        self.connect("page-added", self.__page_added)
        self.connect("page-removed", self.__page_removed)
        self.connect("create-window", self.__create_window)

        config_settings, opts_list = self.__load(defaults)

        opts_list.sort(key=lambda opt: opt["page"])
        for opts in opts_list:
            if opts["name"] in AVAILABLE_GRAMPLETS():
                all_opts = get_gramplet_opts(opts["name"], opts)
                gramplet = make_requested_gramplet(
                    TabGramplet, self, all_opts, self.dbstate, self.uistate
                )
                if gramplet:
                    self.__add_tab(gramplet)

        if len(opts_list) == 0:
            self.empty = True
            self.__create_empty_tab()

        if config_settings[0]:
            self.show()
        self.set_current_page(config_settings[1])

        uistate.connect("grampletbar-close-changed", self.cb_close_changed)

        # Connect after gramplets added to prevent making them active
        self.connect("switch-page", self.__switch_page)

    def _get_config_setting(self, configparser, section, setting, fn=None):
        """
        Get a section.setting value from the config parser.
        Takes a configparser instance, a section, a setting, and
        optionally a post-processing function (typically int).

        Always returns a value of the appropriate type.
        """
        value = ""
        try:
            value = configparser.get(section, setting)
            value = value.strip()
            if fn:
                value = fn(value)
        except:
            if fn:
                value = fn()
            else:
                value = ""
        return value

    def __load(self, defaults):
        """
        Load the gramplets from the configuration file.
        """
        retval = []
        visible = True
        default_page = 0
        filename = self.configfile
        if filename and os.path.exists(filename):
            cp = configparser.ConfigParser()
            try:
                cp.read(filename, encoding="utf-8")
            except:
                pass
            for sec in cp.sections():
                if sec == "Bar Options":
                    if "visible" in cp.options(sec):
                        visible = self._get_config_setting(cp, sec, "visible") == "True"
                    if "page" in cp.options(sec):
                        default_page = self._get_config_setting(cp, sec, "page", int)
                else:
                    data = {}
                    for opt in cp.options(sec):
                        if opt.startswith("data["):
                            temp = data.get("data", {})
                            # temp.append(self._get_config_setting(cp, sec, opt))
                            pos = int(opt[5:-1])
                            temp[pos] = self._get_config_setting(cp, sec, opt)
                            data["data"] = temp
                        else:
                            data[opt] = self._get_config_setting(cp, sec, opt)
                    if "data" in data:
                        data["data"] = [
                            data["data"][key] for key in sorted(data["data"].keys())
                        ]
                    if "name" not in data:
                        data["name"] = "Unnamed Gramplet"
                        data["tname"] = _("Unnamed Gramplet")
                    retval.append(data)
            clean_up(cp)
        else:
            # give defaults as currently known
            for name in defaults:
                if name in AVAILABLE_GRAMPLETS():
                    retval.append(GET_AVAILABLE_GRAMPLETS(name))
        return ((visible, default_page), retval)

    def __save(self):
        """
        Save the gramplet configuration.
        """
        filename = self.configfile
        try:
            with open(filename, "w", encoding="utf-8") as fp:
                fp.write(";; Gramplet bar configuration file" + NL)
                fp.write(
                    (
                        ";; Automatically created at %s"
                        % time.strftime("%Y/%m/%d %H:%M:%S")
                    )
                    + NL
                    + NL
                )
                fp.write("[Bar Options]" + NL)
                fp.write(("visible=%s" + NL) % self.get_property("visible"))
                fp.write(("page=%d" + NL) % self.get_current_page())
                fp.write(NL)

                if self.empty:
                    gramplet_list = []
                else:
                    gramplet_list = [
                        self.get_nth_page(page_num)
                        for page_num in range(self.get_n_pages())
                    ]

                for page_num, gramplet in enumerate(gramplet_list):
                    opts = get_gramplet_options_by_name(gramplet.gname)
                    if opts is not None:
                        base_opts = opts.copy()
                        for key in base_opts:
                            if key in gramplet.__dict__:
                                base_opts[key] = gramplet.__dict__[key]
                        fp.write(("[%s]" + NL) % gramplet.gname)
                        for key in base_opts:
                            if key in [
                                "content",
                                "title",
                                "tname",
                                "row",
                                "column",
                                "page",
                                "version",
                                "gramps",
                            ]:  # don't save
                                continue
                            elif key == "data":
                                if not isinstance(base_opts["data"], (list, tuple)):
                                    fp.write(("data[0]=%s" + NL) % base_opts["data"])
                                else:
                                    cnt = 0
                                    for item in base_opts["data"]:
                                        fp.write(("data[%d]=%s" + NL) % (cnt, item))
                                        cnt += 1
                            else:
                                fp.write(("%s=%s" + NL) % (key, base_opts[key]))
                        fp.write(("page=%d" + NL) % page_num)
                        fp.write(NL)

        except IOError:
            LOG.warning("Failed writing '%s'; gramplets not saved" % filename)
            return

    def set_active(self):
        """
        Called with the view is set as active.
        """
        if not self.empty:
            gramplet = self.get_nth_page(self.get_current_page())
            if gramplet and gramplet.pui:
                gramplet.pui.active = True
                if gramplet.pui.dirty:
                    gramplet.pui.update()

    def set_inactive(self):
        """
        Called with the view is set as inactive.
        """
        if not self.empty:
            gramplet = self.get_nth_page(self.get_current_page())
            if gramplet and gramplet.pui:
                gramplet.pui.active = False

    def on_delete(self):
        """
        Called when the view is closed.
        """
        list(map(self.__dock_gramplet, self.detached_gramplets))
        if not self.empty:
            for page_num in range(self.get_n_pages()):
                gramplet = self.get_nth_page(page_num)
                # this is the only place where the gui runs user code directly
                if gramplet.pui:
                    gramplet.pui.on_save()
        self.__save()

    def add_gramplet(self, gname):
        """
        Add a gramplet by name.
        """
        if self.has_gramplet(gname):
            return
        all_opts = get_gramplet_options_by_name(gname)
        gramplet = make_requested_gramplet(
            TabGramplet, self, all_opts, self.dbstate, self.uistate
        )
        if not gramplet:
            LOG.warning("Problem creating '%s'", gname)
            return

        page_num = self.__add_tab(gramplet)
        self.set_current_page(page_num)

    def remove_gramplet(self, gname):
        """
        Remove a gramplet by name.
        """
        for gramplet in self.detached_gramplets:
            if gramplet.gname == gname:
                self.__dock_gramplet(gramplet)
                self.remove_page(self.page_num(gramplet))
                return

        for page_num in range(self.get_n_pages()):
            gramplet = self.get_nth_page(page_num)
            if gramplet.gname == gname:
                self.remove_page(page_num)
                return

    def has_gramplet(self, gname):
        """
        Return True if the GrampletBar contains the gramplet, else False.
        """
        return gname in self.all_gramplets()

    def all_gramplets(self):
        """
        Return a list of names of all the gramplets in the GrampletBar.
        """
        if self.empty:
            return self.detached_gramplets
        else:
            return [
                gramplet.gname
                for gramplet in self.get_children() + self.detached_gramplets
            ]

    def restore(self):
        """
        Restore the GrampletBar to its default gramplets.
        """
        list(map(self.remove_gramplet, self.all_gramplets()))
        list(map(self.add_gramplet, self.defaults))
        self.set_current_page(0)

    def __create_empty_tab(self):
        """
        Create an empty tab to be displayed when the GrampletBar is empty.
        """
        tab_label = Gtk.Label(label=_("Gramplet Bar"))
        tab_label.show()
        msg = _(
            "Select the down arrow on the right corner for adding, removing or restoring gramplets."
        )
        content = Gtk.Label(label=msg)
        content.set_halign(Gtk.Align.START)
        content.set_line_wrap(True)
        content.set_size_request(150, -1)
        content.show()
        self.append_page(content, tab_label)
        return content

    def __add_tab(self, gramplet):
        """
        Add a tab to the notebook for the given gramplet.
        """
        label = self.__create_tab_label(gramplet)
        page_num = self.append_page(gramplet, label)
        return page_num

    def __create_tab_label(self, gramplet):
        """
        Create a tab label consisting of a label and a close button.
        """
        tablabel = TabLabel(gramplet, self.__delete_clicked)

        if hasattr(gramplet.pui, "has_data"):
            tablabel.set_has_data(gramplet.pui.has_data)
        else:  # just a function; always show yes it has data
            tablabel.set_has_data(True)

        if config.get("interface.grampletbar-close"):
            tablabel.use_close(True)
        else:
            tablabel.use_close(False)

        return tablabel

    def cb_close_changed(self):
        """
        Close button preference changed.
        """
        for gramplet in self.get_children():
            tablabel = self.get_tab_label(gramplet)
            if not isinstance(tablabel, Gtk.Label):
                tablabel.use_close(config.get("interface.grampletbar-close"))

    def __delete_clicked(self, button, gramplet):
        """
        Called when the delete button is clicked.
        """
        page_num = self.page_num(gramplet)
        self.remove_page(page_num)

    def __switch_page(self, notebook, unused, new_page):
        """
        Called when the user has switched to a new GrampletBar page.
        """
        old_page = notebook.get_current_page()
        if old_page >= 0:
            gramplet = self.get_nth_page(old_page)
            if gramplet and gramplet.pui:
                gramplet.pui.active = False

        gramplet = self.get_nth_page(new_page)
        if not self.empty:
            if gramplet and gramplet.pui:
                gramplet.pui.active = True
                if gramplet.pui.dirty:
                    gramplet.pui.update()

    def __page_added(self, notebook, unused, new_page):
        """
        Called when a new page is added to the GrampletBar.
        """
        gramplet = self.get_nth_page(new_page)
        if isinstance(gramplet, TabGramplet):
            gramplet.set_orientation(self.orientation)
        if self.empty:
            if isinstance(gramplet, TabGramplet):
                self.empty = False
                if new_page == 0:
                    self.remove_page(1)
                else:
                    self.remove_page(0)
            else:
                return
        gramplet.pane = self
        label = self.__create_tab_label(gramplet)
        self.set_tab_label(gramplet, label)
        self.set_tab_reorderable(gramplet, True)
        self.set_tab_detachable(gramplet, True)
        if gramplet in self.detached_gramplets:
            self.detached_gramplets.remove(gramplet)
            self.reorder_child(gramplet, gramplet.page)

    def __page_removed(self, notebook, unused, page_num):
        """
        Called when a page is removed to the GrampletBar.
        """
        if self.get_n_pages() == 0:
            self.empty = True
            self.__create_empty_tab()

    def __create_window(self, grampletbar, gramplet, x_pos, y_pos):
        """
        Called when the user has switched to a new GrampletBar page.
        """
        gramplet.page = self.page_num(gramplet)
        self.detached_gramplets.append(gramplet)
        win = DetachedWindow(grampletbar, gramplet, x_pos, y_pos)
        gramplet.detached_window = win
        return win.get_notebook()

    def __dock_gramplet(self, gramplet):
        """
        Dock a detached gramplet.
        """
        gramplet.detached_window.close()
        gramplet.detached_window = None

    def __button_clicked(self, button):
        """
        Called when the drop-down button is clicked.
        """
        self.menu = Gtk.Menu()
        menu = self.menu

        ag_menu = Gtk.MenuItem(label=_("Add a gramplet"))
        nav_type = self.pageview.navigation_type()
        skip = self.all_gramplets()
        gramplet_list = GET_GRAMPLET_LIST(nav_type, skip)
        gramplet_list.sort()
        self.__create_submenu(ag_menu, gramplet_list, self.__add_clicked)
        ag_menu.show()
        menu.append(ag_menu)

        if not (self.empty or config.get("interface.grampletbar-close")):
            rg_menu = Gtk.MenuItem(label=_("Remove a gramplet"))
            gramplet_list = [
                (gramplet.title, gramplet.gname)
                for gramplet in self.get_children() + self.detached_gramplets
            ]
            gramplet_list.sort()
            self.__create_submenu(rg_menu, gramplet_list, self.__remove_clicked)
            rg_menu.show()
            menu.append(rg_menu)

        rd_menu = Gtk.MenuItem(label=_("Restore default gramplets"))
        rd_menu.connect("activate", self.__restore_clicked)
        rd_menu.show()
        menu.append(rd_menu)

        # Separator.
        rs_menu = Gtk.SeparatorMenuItem()
        rs_menu.show()
        menu.append(rs_menu)

        rh_menu = Gtk.MenuItem(label=_("Gramplet Bar Help"))
        rh_menu.connect("activate", self.on_help_grampletbar_clicked)
        rh_menu.show()
        menu.append(rh_menu)

        rg_menu = Gtk.MenuItem(label=_("About Gramplets"))
        rg_menu.connect("activate", self.on_help_gramplets_clicked)
        rg_menu.show()
        menu.append(rg_menu)

        menu.show_all()
        menu.popup_at_widget(
            button, Gdk.Gravity.SOUTH_WEST, Gdk.Gravity.NORTH_WEST, None
        )

    def __create_submenu(self, main_menu, gramplet_list, callback_func):
        """
        Create a submenu of the context menu.
        """
        if main_menu:
            submenu = main_menu.get_submenu()
            submenu = Gtk.Menu()
            for entry in gramplet_list:
                item = Gtk.MenuItem(label=entry[0])
                item.connect("activate", callback_func, entry[1])
                item.show()
                submenu.append(item)
            main_menu.set_submenu(submenu)

    def __add_clicked(self, menu, gname):
        """
        Called when a gramplet is added from the context menu.
        """
        self.add_gramplet(gname)

    def __remove_clicked(self, menu, gname):
        """
        Called when a gramplet is removed from the context menu.
        """
        self.remove_gramplet(gname)

    def __restore_clicked(self, menu):
        """
        Called when restore defaults is clicked from the context menu.
        """
        QuestionDialog(
            _("Restore to defaults?"),
            _(
                "The gramplet bar will be restored to contain its default "
                "gramplets.  This action cannot be undone."
            ),
            _("OK"),
            self.restore,
            parent=self.uistate.window,
        )

    def get_config_funcs(self):
        """
        Return a list of configuration functions.
        """
        funcs = []
        if self.empty:
            gramplets = []
        else:
            gramplets = self.get_children()
        for gramplet in gramplets + self.detached_gramplets:
            gui_options = gramplet.make_gui_options()
            if gui_options:
                funcs.append(self.__build_panel(gramplet.title, gui_options))
        return funcs

    def __build_panel(self, title, gui_options):
        """
        Return a configuration function that returns the title of a page in
        the Configure View dialog and a gtk container defining the page.
        """

        def gramplet_panel(configdialog):
            return title, gui_options

        return gramplet_panel

    def on_help_grampletbar_clicked(self, dummy):
        """Button: Display the relevant portion of Gramps manual"""
        display_help(MAIN_HELP_PAGE, "Bottombar_and_Sidebar")

    def on_help_gramplets_clicked(self, dummy):
        """Button: Display the relevant portion of Gramps manual"""
        display_help(WIKI_HELP_PAGE, "What_is_a_Gramplet?")


# -------------------------------------------------------------------------
#
# TabGramplet class
#
# -------------------------------------------------------------------------
class TabGramplet(Gtk.ScrolledWindow, GuiGramplet):
    """
    Class that handles the plugin interfaces for the GrampletBar.
    """

    def __init__(self, pane, dbstate, uistate, title, **kwargs):
        """
        Internal constructor for GUI portion of a gramplet.
        """
        Gtk.ScrolledWindow.__init__(self)
        GuiGramplet.__init__(self, pane, dbstate, uistate, title, **kwargs)

        self.scrolledwindow = self
        self.textview = Gtk.TextView()
        self.textview.set_editable(False)
        self.textview.set_wrap_mode(Gtk.WrapMode.WORD)
        self.buffer = UndoableBuffer()
        self.text_length = 0
        self.textview.set_buffer(self.buffer)
        self.textview.connect("key-press-event", self.on_key_press_event)
        self.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.add(self.textview)
        self.show_all()
        self.track = []

    def get_title(self):
        return self.title

    def get_container_widget(self):
        """
        Return the top level container widget.
        """
        return self

    def set_orientation(self, orientation):
        """
        Called when the gramplet orientation changes.
        """
        self.pui.set_orientation(orientation)


# -------------------------------------------------------------------------
#
# DetachedWindow class
#
# -------------------------------------------------------------------------
class DetachedWindow(ManagedWindow):
    """
    Class for showing a detached gramplet.
    """

    def __init__(self, grampletbar, gramplet, x_pos, y_pos):
        """
        Construct the window.
        """
        self.title = gramplet.title + " " + _("Gramplet")
        self.grampletbar = grampletbar
        self.gramplet = gramplet

        ManagedWindow.__init__(self, gramplet.uistate, [], self.title)
        dlg = Gtk.Dialog(
            transient_for=gramplet.uistate.window, destroy_with_parent=True
        )
        dlg.add_button(_("_Close"), Gtk.ResponseType.CLOSE)
        self.set_window(dlg, None, self.title)
        self.window.move(x_pos, y_pos)
        self.window.set_default_size(gramplet.detached_width, gramplet.detached_height)
        self.window.add_button(_("_Help"), Gtk.ResponseType.HELP)
        self.window.connect("response", self.handle_response)

        self.notebook = Gtk.Notebook()
        self.notebook.set_show_tabs(False)
        self.notebook.set_show_border(False)
        self.notebook.connect("page-added", self.page_added)
        self.notebook.show()
        self.window.vbox.pack_start(self.notebook, True, True, 0)
        self.show()

    def page_added(self, notebook, gramplet, page_num):
        """
        Called when the gramplet is added to the notebook.  This takes the
        focus from the help button (bug #6306).
        """
        gramplet.grab_focus()

    def handle_response(self, object, response):
        """
        Callback for taking care of button clicks.
        """
        if response == Gtk.ResponseType.CLOSE:
            self.close()
        elif response == Gtk.ResponseType.HELP:
            # translated name:
            if self.gramplet.help_url:
                if self.gramplet.help_url.startswith(("http://", "https://")):
                    display_url(self.gramplet.help_url)
                else:
                    display_help(self.gramplet.help_url)
            else:
                display_help(WIKI_HELP_PAGE, self.gramplet.tname.replace(" ", "_"))

    def get_notebook(self):
        """
        Return the notebook.
        """
        return self.notebook

    def build_menu_names(self, obj):
        """
        Part of the Gramps window interface.
        """
        return (self.title, "Gramplet")

    def get_title(self):
        """
        Returns the window title.
        """
        return self.title

    def close(self, *args):
        """
        Dock the detached gramplet back in the GrampletBar from where it came.
        """
        size = self.window.get_size()
        self.gramplet.detached_width = size[0]
        self.gramplet.detached_height = size[1]
        self.gramplet.detached_window = None
        self.notebook.remove(self.gramplet)
        self.grampletbar.add(self.gramplet)
        ManagedWindow.close(self, *args)


# -------------------------------------------------------------------------
#
# TabLabel class
#
# -------------------------------------------------------------------------
class TabLabel(Gtk.Box):
    """
    Create a tab label consisting of a label and a close button.
    """

    def __init__(self, gramplet, callback):
        Gtk.Box.__init__(self)

        self.text = gramplet.title
        self.set_spacing(4)

        self.label = Gtk.Label()
        self.label.set_tooltip_text(gramplet.tname)
        self.label.show()

        self.closebtn = Gtk.Button()
        image = Gtk.Image()
        image.set_from_icon_name("window-close", Gtk.IconSize.MENU)
        self.closebtn.connect("clicked", callback, gramplet)
        self.closebtn.set_image(image)
        self.closebtn.set_relief(Gtk.ReliefStyle.NONE)

        self.pack_start(self.label, True, True, 0)
        self.pack_end(self.closebtn, False, False, 0)

    def set_has_data(self, has_data):
        """
        Set the label to indicate if the gramplet has data.
        """
        if has_data:
            self.label.set_text("<b>%s</b>" % self.text)
            self.label.set_use_markup(True)
        else:
            self.label.set_text(self.text)

    def use_close(self, use_close):
        """
        Display the cose button according to user preference.
        """
        if use_close:
            self.closebtn.show()
        else:
            self.closebtn.hide()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2005  Donald N. Allingham
# Copyright (C) 2012       Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Gender statistics kept in Gramps database.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .person import Person


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
class GenderStats:
    """
    Class for keeping track of statistics related to Given Name vs. Gender.

    This allows the tracking of the liklihood of a person's given name
    indicating the gender of the person.
    """

    def __init__(self, stats=None):
        if stats is None:
            self.stats = {}
        else:
            # Current use of GenderStats is such that a shallow copy suffices.
            self.stats = stats

    def save_stats(self):
        return self.stats

    def clear_stats(self):
        self.stats = {}
        return self.stats

    def name_stats(self, name):
        if name in self.stats:
            return self.stats[name]
        return (0, 0, 0)

    def count_name(self, name, gender):
        """
        Count a given name under gender in the gender stats.
        """
        keyname = _get_key_from_name(name)
        if not keyname:
            return

        self._set_stats(keyname, gender)

    def count_person(self, person, undo=0):
        if not person:
            return
        # Let the Person do their own counting later

        keyname = _get_key(person)
        if not keyname:
            return

        gender = person.get_gender()
        self._set_stats(keyname, gender, undo)

    def _set_stats(self, keyname, gender, undo=0):
        (male, female, unknown) = self.name_stats(keyname)
        if not undo:
            increment = 1
        else:
            increment = -1

        if gender == Person.MALE:
            male += increment
            if male < 0:
                male = 0
        elif gender == Person.FEMALE:
            female += increment
            if female < 0:
                female = 0
        elif gender in (Person.UNKNOWN, Person.OTHER):
            unknown += increment
            if unknown < 0:
                unknown = 0

        self.stats[keyname] = (male, female, unknown)

    def uncount_person(self, person):
        return self.count_person(person, undo=1)

    def guess_gender(self, name):
        name = _get_key_from_name(name)
        if not name or name not in self.stats:
            return Person.UNKNOWN

        (male, female, unknown) = self.stats[name]
        if unknown == 0:
            if male and not female:
                return Person.MALE
            if female and not male:
                return Person.FEMALE

        if male > (2 * female):
            return Person.MALE

        if female > (2 * male):
            return Person.FEMALE

        return Person.UNKNOWN


def _get_key(person):
    name = person.get_primary_name().get_first_name()
    return _get_key_from_name(name)


def _get_key_from_name(name):
    return name.split(" ")[0].replace("?", "")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2006  Donald N. Allingham
# Copyright (C) 2023       Avi Markovitz
# Copyright (C) 2023       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Hebrew-specific classes for parsing and displaying dates.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import re

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib.date import Date
from ._dateparser import DateParser
from ._datedisplay import DateDisplay
from ._datehandler import register_datehandler


# -------------------------------------------------------------------------
#
# Hebrew parser
#
# -------------------------------------------------------------------------


class DateParserHE(DateParser):
    calendar_to_int = {
        "גרגוריאני": Date.CAL_GREGORIAN,
        "אזרחי": Date.CAL_GREGORIAN,
        "יוליאני": Date.CAL_JULIAN,
        "י": Date.CAL_JULIAN,
        "עברי": Date.CAL_HEBREW,
        "ע": Date.CAL_HEBREW,
        "מוסלמי": Date.CAL_ISLAMIC,
        "מ": Date.CAL_ISLAMIC,
        "המהפכה הצרפתית": Date.CAL_FRENCH,
        "צ": Date.CAL_FRENCH,
        "פרסי": Date.CAL_PERSIAN,
        "פ": Date.CAL_PERSIAN,
        "שוודי": Date.CAL_SWEDISH,
        "ש": Date.CAL_SWEDISH,
    }

    # Where there are substrings, put the longest string first. e.g. "ב־" before "ב".
    modifier_to_int = {
        "ב־": Date.MOD_NONE,
        "ב": Date.MOD_NONE,
        "לפני": Date.MOD_BEFORE,
        "לפני ה־": Date.MOD_BEFORE,
        "לפ.": Date.MOD_BEFORE,
        "אחרי": Date.MOD_AFTER,
        "אחרי ה־": Date.MOD_AFTER,
        "אח.": Date.MOD_AFTER,
        "בסביבות־": Date.MOD_ABOUT,
        "בסביבות": Date.MOD_ABOUT,
        "בערך ב־": Date.MOD_ABOUT,
        "בערך ב": Date.MOD_ABOUT,
        "באזור ה־": Date.MOD_ABOUT,
        "באזור ה": Date.MOD_ABOUT,
        "קרוב ל־": Date.MOD_ABOUT,
        "קרוב ל": Date.MOD_ABOUT,
        "בקרוב": Date.MOD_ABOUT,
        "במקורב": Date.MOD_ABOUT,
        "מיום": Date.MOD_FROM,
        "מה־": Date.MOD_FROM,
        "מ־": Date.MOD_FROM,
        "מ": Date.MOD_FROM,
        "עד ל־": Date.MOD_TO,
        "עד ל": Date.MOD_TO,
        "עד": Date.MOD_TO,
        "עד יום": Date.MOD_TO,
        "עד ה־": Date.MOD_TO,
        "ועד יום": Date.MOD_TO,
    }

    quality_to_int = {
        "מוערך": Date.QUAL_ESTIMATED,
        "משוער": Date.QUAL_ESTIMATED,
        "מחושב": Date.QUAL_CALCULATED,
    }

    bce = [
        "לפני הספירה",
        "לפני עידן זה",
        'לפנה"\\ס',
        "לפני ספירת הנוצרים",
        "לספירתם",
    ] + DateParser.bce

    def init_strings(self):
        DateParser.init_strings(self)
        self._modifier = re.compile(r"%s\s*(.*)" % self._mod_str, re.IGNORECASE)
        self._text2 = re.compile(
            r"(\d+)?\s+?ב?%s\.?\s*((\d+)(/\d+)?)?\s*$" % self._mon_str, re.IGNORECASE
        )
        self._jtext2 = re.compile(
            r"(\d+)?\s+?ב?%s\s*((\d+)(/\d+)?)?\s*$" % self._jmon_str, re.IGNORECASE
        )
        self._ftext2 = re.compile(
            r"(\d+)?\s+?ב?%s\s*((\d+)(/\d+)?)?\s*$" % self._fmon_str, re.IGNORECASE
        )
        self._ptext2 = re.compile(
            r"(\d+)?\s+?ב?%s\s*((\d+)(/\d+)?)?\s*$" % self._pmon_str, re.IGNORECASE
        )
        self._itext2 = re.compile(
            r"(\d+)?\s+?ב?%s\s*((\d+)(/\d+)?)?\s*$" % self._imon_str, re.IGNORECASE
        )
        self._stext2 = re.compile(
            r"(\d+)?\s+?ב?%s\.?\s*((\d+)(/\d+)?)?\s*$" % self._smon_str, re.IGNORECASE
        )
        _span_1 = ["מ־", "מ"]
        _span_2 = ["עד"]
        _range_1 = ["בין"]
        _range_2 = ["ל־", "ל"]
        self._span = re.compile(
            r"(%s)\s*(?P<start>.+)\s+(%s)\s+(?P<stop>.+)"
            % ("|".join(_span_1), "|".join(_span_2)),
            re.IGNORECASE,
        )
        self._range = re.compile(
            r"(%s)\s+(?P<start>.+)\s+(%s)\s*(?P<stop>.+)"
            % ("|".join(_range_1), "|".join(_range_2)),
            re.IGNORECASE,
        )


# -------------------------------------------------------------------------
#
# Hebrew display
#
# -------------------------------------------------------------------------
class DateDisplayHE(DateDisplay):
    """
    Hebrew language date display class.
    """

    _bce_str = "%s לספירה"

    long_months = (
        "",
        "ינואר",
        "פברואר",
        "מרץ",
        "אפריל",
        "מאי",
        "יוני",
        "יולי",
        "אוגוסט",
        "ספטמבר",
        "אוקטובר",
        "נובמבר",
        "דצמבר",
    )

    short_months = (
        "",
        "ינו",
        "פבר",
        "מרץ",
        "אפר",
        "מאי",
        "יונ",
        "יול",
        "אוג",
        "ספט",
        "אוק",
        "נוב",
        "דצמ",
    )

    hebrew = (
        "",
        "תשרי",
        "חשוון",
        "כסלו",
        "תבט",
        "שבט",
        "אדר",
        "אדר א'",
        "ניסן",
        "אייר",
        "סיוון",
        "תמוז",
        "אב",
        "אלול",
    )

    formats = (
        "DD-MM-AAAA (ISO)",
        "סיפרתי",
        "חודש יום, שנה",
        "חודש יום, שנה",
        "יום חודש, שנה",
        "יום חודש, שנה",
    )
    # this must agree with DateDisplayEn's "formats" definition
    # (since no locale-specific _display_gregorian exists, here)

    def display(self, date):
        """
        Return a text string representing the date.
        """
        mod = date.get_modifier()
        cal = date.get_calendar()
        qual = date.get_quality()
        start = date.get_start_date()
        newyear = date.get_new_year()

        qual_str = self._qual_str[qual]

        if mod == Date.MOD_TEXTONLY:
            return date.get_text()
        elif start == Date.EMPTY:
            return ""
        elif mod == Date.MOD_SPAN:
            d1 = self.display_cal[cal](start)
            d2 = self.display_cal[cal](date.get_stop_date())
            scal = self.format_extras(cal, newyear)
            return "%s%s %s %s%s" % (qual_str, add_prefix(d1, "מ"), "עד", d2, scal)
        elif mod == Date.MOD_RANGE:
            d1 = self.display_cal[cal](start)
            d2 = self.display_cal[cal](date.get_stop_date())
            scal = self.format_extras(cal, newyear)
            return "%s%s %s %s%s" % (qual_str, "בין", d1, add_prefix(d2, "ל"), scal)
        elif mod == Date.MOD_NONE:
            text = self.display_cal[cal](start)
            scal = self.format_extras(cal, newyear)
            if start[0] == 0 or qual != Date.QUAL_NONE:
                text = add_prefix(text, "ב")
            return "%s%s%s" % (qual_str, text, scal)
        else:
            text = self.display_cal[date.get_calendar()](start)
            scal = self.format_extras(cal, newyear)
            term = self._mod_str[mod]
            if term.endswith(" "):
                modifier = term + text
            else:
                modifier = add_prefix(text, term)
            return "%s%s%s" % (qual_str, modifier, scal)


def add_prefix(text, prefix):
    """
    Return a prefixed string with a maqaf added for non-Hebrew text.
    """
    if text[0] < "א" or text[0] > "ת":
        return prefix + "־" + text
    else:
        return prefix + text


# -------------------------------------------------------------------------
#
# Register classes
#
# -------------------------------------------------------------------------
register_datehandler(
    ("he_IL", "he", "Hebrew", "Ivrit", "עברית", ("%d-%m-%Y",)),
    DateParserHE,
    DateDisplayHE,
)



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011 Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Gtk modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gui.listmodel import ListModel
from gramps.gui.plug.quick import run_quick_report_by_name
from gramps.gen.plug import Gramplet
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


class Attributes(Gramplet):
    """
    Displays the attributes of an object.
    """

    def init(self):
        self.gui.WIDGET = self.build_gui()
        self.gui.get_container_widget().remove(self.gui.textview)
        self.gui.get_container_widget().add(self.gui.WIDGET)
        self.gui.WIDGET.show()

    def build_gui(self):
        """
        Build the GUI interface.
        """
        tip = _(
            "Double-click on a row to view a quick report showing "
            "all people with the selected attribute."
        )
        self.set_tooltip(tip)
        top = Gtk.TreeView()
        titles = [(_("Key"), 1, 100), (_("Value"), 2, 100)]
        self.model = ListModel(top, titles, event_func=self.display_report)
        return top

    def display_attributes(self, obj):
        """
        Display the attributes of an object.
        """
        for attr in obj.get_attribute_list():
            self.model.add((str(attr.get_type()), attr.get_value()))
        self.set_has_data(self.model.count > 0)

    def display_report(self, treeview):
        """
        Display the quick report for matching attribute key.
        """
        model, iter_ = treeview.get_selection().get_selected()
        if iter_:
            key = model.get_value(iter_, 0)
            run_quick_report_by_name(self.dbstate, self.uistate, "attribute_match", key)

    def get_has_data(self, obj):
        """
        Return True if the gramplet has data, else return False.
        """
        if obj is None:
            return False
        if obj.get_attribute_list():
            return True
        return False


class PersonAttributes(Attributes):
    """
    Displays the attributes of a person.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "person-update", self.update)

    def active_changed(self, handle):
        self.update()

    def update_has_data(self):
        active_handle = self.get_active("Person")
        if active_handle:
            active = self.dbstate.db.get_person_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.model.clear()
        active_handle = self.get_active("Person")
        if active_handle:
            active = self.dbstate.db.get_person_from_handle(active_handle)
            if active:
                self.display_attributes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)


class EventAttributes(Attributes):
    """
    Displays the attributes of an event.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "event-update", self.update)
        self.connect_signal("Event", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Event")
        if active_handle:
            active = self.dbstate.db.get_event_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.model.clear()
        active_handle = self.get_active("Event")
        if active_handle:
            active = self.dbstate.db.get_event_from_handle(active_handle)
            if active:
                self.display_attributes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)


class FamilyAttributes(Attributes):
    """
    Displays the attributes of an event.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "family-update", self.update)
        self.connect_signal("Family", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Family")
        if active_handle:
            active = self.dbstate.db.get_family_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.model.clear()
        active_handle = self.get_active("Family")
        if active_handle:
            active = self.dbstate.db.get_family_from_handle(active_handle)
            if active:
                self.display_attributes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)


class MediaAttributes(Attributes):
    """
    Displays the attributes of a media object.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "media-update", self.update)
        self.connect_signal("Media", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Media")
        if active_handle:
            active = self.dbstate.db.get_media_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.model.clear()
        active_handle = self.get_active("Media")
        if active_handle:
            active = self.dbstate.db.get_media_from_handle(active_handle)
            if active:
                self.display_attributes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)


class SourceAttributes(Attributes):
    """
    Displays the attributes of a source object.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "source-update", self.update)
        self.connect_signal("Source", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Source")
        if active_handle:
            active = self.dbstate.db.get_source_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.model.clear()
        active_handle = self.get_active("Source")
        if active_handle:
            active = self.dbstate.db.get_source_from_handle(active_handle)
            if active:
                self.display_attributes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)


class CitationAttributes(Attributes):
    """
    Displays the attributes of a citation object.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "citation-update", self.update)
        self.connect_signal("Citation", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Citation")
        if active_handle:
            active = self.dbstate.db.get_citation_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.model.clear()
        active_handle = self.get_active("Citation")
        if active_handle:
            active = self.dbstate.db.get_citation_from_handle(active_handle)
            if active:
                self.display_attributes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2022       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Handles generation and access to thumbnails used in Gramps.
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import os
import logging
from hashlib import md5

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import GLib
from gi.repository import GdkPixbuf

try:
    from gi.repository import Gtk

    _icon_theme = Gtk.IconTheme.get_default()
except:
    _icon_theme = None

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import (
    ICON,
    IMAGE_DIR,
    THUMB_LARGE,
    THUMB_NORMAL,
    SIZE_NORMAL,
    SIZE_LARGE,
)
from gramps.gen.plug import BasePluginManager, START
from gramps.gen.mime import get_type

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
LOG = logging.getLogger(".thumbnail")

THUMBNAILERS = []


def get_thumbnailers():
    if len(THUMBNAILERS):
        return THUMBNAILERS

    plugman = BasePluginManager.get_instance()
    for pdata in plugman.get_reg_thumbnailers():
        module = plugman.load_plugin(pdata)
        if not module:
            print("Error loading thumbnailer '%s': skipping content" % pdata.name)
            continue
        thumbnailer = getattr(module, pdata.thumbnailer)()
        if pdata.order == START:
            THUMBNAILERS.insert(0, thumbnailer)
        else:
            THUMBNAILERS.append(thumbnailer)
    return THUMBNAILERS


# -------------------------------------------------------------------------
#
# __build_thumb_path
#
# -------------------------------------------------------------------------
def __build_thumb_path(path, rectangle=None, size=SIZE_NORMAL):
    """
    Convert the specified path into a corresponding path for the thumbnail
    image. We do this by converting the original path into an MD5SUM value
    (which should be unique), adding the '.png' extension, and prepending
    with the Gramps thumbnail directory.

    :type path: unicode
    :param path: filename of the source file
    :type rectangle: tuple
    :param rectangle: subsection rectangle
    :rtype: unicode
    :returns: full path name to the corresponding thumbnail file.
    """
    extra = ""
    if rectangle is not None:
        extra = "?" + str(rectangle)
    prehash = path + extra
    prehash = prehash.encode("utf-8")
    md5_hash = md5(prehash)
    if size == SIZE_LARGE:
        base_dir = THUMB_LARGE
    else:
        base_dir = THUMB_NORMAL
    return os.path.join(base_dir, md5_hash.hexdigest() + ".png")


# -------------------------------------------------------------------------
#
# __create_thumbnail_image
#
# -------------------------------------------------------------------------
def __create_thumbnail_image(src_file, mtype=None, rectangle=None, size=SIZE_NORMAL):
    """
    Generates the thumbnail image for a file. If the mime type is specified,
    and is not an 'image', then we attempt to find and run a thumbnailer
    utility to create a thumbnail. For images, we simply create a smaller
    image, scaled to thumbnail size.

    :param src_file: filename of the source file
    :type src_file: unicode
    :param mtype: mime type of the specified file (optional)
    :type mtype: unicode
    :param rectangle: subsection rectangle
    :type rectangle: tuple
    :rtype: bool
    :returns: True is the thumbnailwas successfully generated
    """
    if mtype is None:
        mtype = get_type(src_file)
    filename = __build_thumb_path(src_file, rectangle, size)
    return run_thumbnailer(mtype, src_file, filename, size, rectangle)


def run_thumbnailer(mime_type, src_file, dest_file, size, rectangle=None):
    """
    This function attempts to generate a thumbnail image.  It runs the first
    thumbnailer plugin that supports the given mime type.

    :param mime_type: mime type of the source file
    :type mime_type: unicode
    :param src_file: filename of the source file
    :type src_file: unicode
    :param dest_file: destination file for the thumbnail image
    :type dest_file: unicode
    :param size: option parameters specifying the desired size of the
      thumbnail
    :type size: int
    :param rectangle: subsection rectangle (optional)
    :type rectangle: tuple
    :returns: True if the thumbnail was successfully generated
    :rtype: bool
    """
    for thumbnailer in get_thumbnailers():
        if thumbnailer.is_supported(mime_type):
            if thumbnailer.run(mime_type, src_file, dest_file, size, rectangle):
                return True
    return False


# -------------------------------------------------------------------------
#
# find_mime_type_pixbuf
#
# -------------------------------------------------------------------------


def find_mime_type_pixbuf(mime_type):
    try:
        icontmp = mime_type.replace("/", "-")
        newicon = "gnome-mime-%s" % icontmp
        try:
            return _icon_theme.load_icon(newicon, 48, 0)
        except:
            icontmp = mime_type.split("/")[0]
            try:
                newicon = "gnome-mime-%s" % icontmp
                return _icon_theme.load_icon(newicon, 48, 0)
            except:
                return GdkPixbuf.Pixbuf.new_from_file(ICON)
    except:
        return GdkPixbuf.Pixbuf.new_from_file(ICON)


# -------------------------------------------------------------------------
#
# get_thumbnail_image
#
# -------------------------------------------------------------------------
def get_thumbnail_image(src_file, mtype=None, rectangle=None, size=SIZE_NORMAL):
    """
    Return the thumbnail image (in GTK Pixbuf format) associated with the
    source file passed to the function. If no thumbnail could be found,
    the associated icon for the mime type is returned, or if that cannot be
    found, a generic document icon is returned.

    The image is not generated every time, but only if the thumbnail does not
    exist, or if the source file is newer than the thumbnail.

    :param src_file: Source media file
    :type src_file: unicode
    :param mime_type: mime type of the source file
    :type mime_type: unicode
    :param rectangle: subsection rectangle
    :type rectangle: tuple
    :returns: thumbnail representing the source file
    :rtype: GdkPixbuf.Pixbuf
    """
    try:
        filename = get_thumbnail_path(src_file, mtype, rectangle, size)
        return GdkPixbuf.Pixbuf.new_from_file(filename)
    except (GLib.GError, OSError):
        if mtype:
            return find_mime_type_pixbuf(mtype)
        else:
            default = os.path.join(IMAGE_DIR, "document.png")
            return GdkPixbuf.Pixbuf.new_from_file(default)


# -------------------------------------------------------------------------
#
# get_thumbnail_path
#
# -------------------------------------------------------------------------
def get_thumbnail_path(src_file, mtype=None, rectangle=None, size=SIZE_NORMAL):
    """
    Return the path to the thumbnail image associated with the
    source file passed to the function. If the thumbnail does not exist,
    or if it is older than the source file, we create a new thumbnail image.

    :param src_file: Source media file
    :type src_file: unicode
    :param mime_type: mime type of the source file
    :type mime_type: unicode
    :param rectangle: subsection rectangle
    :type rectangle: tuple
    :returns: thumbnail representing the source file
    :rtype: GdkPixbuf.Pixbuf
    """
    filename = __build_thumb_path(src_file, rectangle, size)
    if not os.path.isfile(src_file):
        return os.path.join(IMAGE_DIR, "image-missing.png")
    else:
        if (not os.path.isfile(filename)) or (
            os.path.getmtime(src_file) > os.path.getmtime(filename)
        ):
            if not __create_thumbnail_image(src_file, mtype, rectangle, size):
                return os.path.join(IMAGE_DIR, "document.png")
        return os.path.abspath(filename)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2006 Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Exports the DbTxn class for managing Gramps transactions and the undo
database.
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import pickle
import logging
from collections import defaultdict
import time
import inspect
import os

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .dbconst import DBLOGNAME

_LOG = logging.getLogger(DBLOGNAME)


# -------------------------------------------------------------------------
#
# Gramps transaction class
#
# -------------------------------------------------------------------------
class DbTxn(defaultdict):
    """
    Define a group of database commits that define a single logical operation.
    """

    __slots__ = (
        "msg",
        "commitdb",
        "db",
        "batch",
        "first",
        "last",
        "timestamp",
        "__dict__",
    )

    def __enter__(self):
        """
        Context manager entry method
        """
        _LOG.debug("    DbTxn %s entered" % hex(id(self)))
        self.start_time = time.time()
        self.db.transaction_begin(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Context manager exit method
        """
        if exc_type is None:
            self.db.transaction_commit(self)
        else:
            self.db.transaction_abort(self)

        elapsed_time = time.time() - self.start_time
        if __debug__ and _LOG.isEnabledFor(logging.DEBUG):
            frame = inspect.currentframe()
            c_frame = frame.f_back
            c_code = c_frame.f_code
            _LOG.debug(
                "    **** DbTxn %s exited. Called from file %s, "
                "line %s, in %s **** %.2f seconds",
                hex(id(self)),
                c_code.co_filename,
                c_frame.f_lineno,
                c_code.co_name,
                elapsed_time,
            )

        return False

    def __init__(self, msg, grampsdb, batch=False, **kwargs):
        """
        Create a new transaction.

        The grampsdb should have transaction_begin/commit/abort methods, and
        a get_undodb method to store undo actions.

        A Transaction instance can be created directly, but it is advised to
        use a context to do this. Like this the user must not worry about
        calling the transaction_xx methods on the database.

        The grampsdb parameter is a reference to the DbWrite object to which
        this transaction will be applied.
        grampsdb.get_undodb() should return a list-like interface that
        stores the commit data. This could be a simple list, or a RECNO-style
        database object.

        The data structure used to handle the transactions (see the add method)
        is a Python dictionary where:

        key = (object type, transaction type) where:
            object type = the numeric type of an object. These are
                          defined as PERSON_KEY = 0, FAMILY_KEY = 1, etc.
                          as imported from dbconst.
            transaction type = a numeric representation of the type of
                          transaction: TXNADD = 0, TXNUPD = 1, TXNDEL = 2

        data = Python list where:
            list element = (handle, data) where:
                handle = handle (database key) of the object in the transaction
                data = pickled representation of the object
        """

        # Conditional on __debug__ because all that frame stuff may be slow
        if __debug__ and _LOG.isEnabledFor(logging.DEBUG):
            caller_frame = inspect.stack()[1]
            # If the call comes from gramps.gen.db.generic.DbGenericTxn.__init__
            # then it is just a dummy redirect, so we need to go back another
            # frame to get any real information. The test does not accurately
            # check this, but seems to be good enough for the current diagnostic
            # purposes.
            if (
                os.path.split(caller_frame[1])[1] == "generic.py"
                and caller_frame[3] == "__init__"
            ):
                caller_frame = inspect.stack()[2]
            _LOG.debug(
                "%sDbTxn %s instantiated for '%s'. Called from file %s, "
                "line %s, in %s"
                % (
                    ("Batch " if batch else "",)
                    + (hex(id(self)),)
                    + (msg,)
                    + (os.path.split(caller_frame[1])[1],)
                    + (tuple(caller_frame[i] for i in range(2, 4)))
                )
            )
        defaultdict.__init__(self, list, {})

        self.msg = msg
        self.commitdb = grampsdb.get_undodb()
        self.db = grampsdb
        self.batch = batch
        for key, value in kwargs.items():
            setattr(self, key, value)
        self.first = None
        self.last = None
        self.timestamp = 0

    def get_description(self):
        """
        Return the text string that describes the logical operation performed
        by the Transaction.
        """
        return self.msg

    def set_description(self, msg):
        """
        Set the text string that describes the logical operation performed by
        the Transaction.
        """
        self.msg = msg

    def add(self, obj_type, trans_type, handle, old_data, new_data):
        """
        Add a commit operation to the Transaction.

        The obj_type is a constant that indicates what type of PrimaryObject
        is being added. The handle is the object's database handle, and the
        data is the tuple returned by the object's serialize method.
        """
        self.last = self.commitdb.append(
            pickle.dumps((obj_type, trans_type, handle, old_data, new_data), 1)
        )
        if self.last is None:
            self.last = len(self.commitdb) - 1
        if self.first is None:
            self.first = self.last
        _LOG.debug("added to trans: %d %d %s" % (obj_type, trans_type, handle))
        self[(obj_type, trans_type)] += [(handle, new_data)]
        return

    def get_recnos(self, reverse=False):
        """
        Return a list of record numbers associated with the transaction.

        While the list is an arbitrary index of integers, it can be used
        to indicate record numbers for a database.
        """

        if self.first is None or self.last is None:
            return []
        if not reverse:
            return range(self.first, self.last + 1)
        else:
            return range(self.last, self.first - 1, -1)

    def get_record(self, recno):
        """
        Return a tuple representing the PrimaryObject type, database handle
        for the PrimaryObject, and a tuple representing the data created by
        the object's serialize method.
        """
        return pickle.loads(self.commitdb[recno])

    def __len__(self):
        """
        Return the number of commits associated with the Transaction.
        """
        if self.first is None or self.last is None:
            return 0
        return self.last - self.first + 1



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# GTK libraries
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from html import escape

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------
from gramps.gen.datehandler import get_date
from gramps.gen.display.name import displayer as name_displayer
from gramps.gen.display.place import displayer as place_displayer
from gramps.gen.utils.string import gender as gender_map
from gramps.gen.lib import EventType
from gramps.gen.utils.db import get_birth_or_fallback, get_death_or_fallback


# -------------------------------------------------------------------------
#
# ChildModel
#
# -------------------------------------------------------------------------
class ChildModel(Gtk.ListStore):
    def __init__(self, child_ref_list, db):
        Gtk.ListStore.__init__(
            self,
            int,
            str,
            str,
            str,
            str,
            str,
            str,
            str,
            str,
            str,
            str,
            str,
            str,
            bool,
            bool,
            object,
        )
        self.db = db
        for index, child_ref in enumerate(child_ref_list):
            child = db.get_person_from_handle(child_ref.ref)
            if child:
                self.append(
                    row=[
                        index + 1,
                        child.get_gramps_id(),
                        name_displayer.display(child),
                        gender_map[child.get_gender()],
                        str(child_ref.get_father_relation()),
                        str(child_ref.get_mother_relation()),
                        self.column_birth_day(child),
                        self.column_death_day(child),
                        self.column_birth_place(child),
                        self.column_death_place(child),
                        name_displayer.sort_string(child.primary_name),
                        self.column_birth_sort(child),
                        self.column_death_sort(child),
                        child_ref.has_citations(),
                        child_ref.get_privacy(),
                        child_ref,
                    ]
                )

    def column_birth_day(self, data):
        birth = get_birth_or_fallback(self.db, data)
        if birth:
            if birth.get_type() == EventType.BIRTH:
                return get_date(birth)
            else:
                return "<i>%s</i>" % escape(get_date(birth))
        else:
            return ""

    def column_birth_sort(self, data):
        """
        Return a sort key to use for the birth column.
        As python int can be larger than C int, we cast int
        to a string of 10 long prepended with 0 as needed.
        This gives correct string sort for years in the millenia around today
        """
        birth = get_birth_or_fallback(self.db, data)
        if birth:
            return "%012d" % birth.get_date_object().get_sort_value()
        else:
            return "%012d" % 0

    def column_death_day(self, data):
        death = get_death_or_fallback(self.db, data)
        if death:
            if death.get_type() == EventType.DEATH:
                return get_date(death)
            else:
                return "<i>%s</i>" % escape(get_date(death))
        else:
            return ""

    def column_death_sort(self, data):
        """
        Return a sort key to use for the death column.
        As python int can be larger than C int, we cast int
        to a string of 10 long prepended with 0 as needed.
        This gives correct string sort for years in the millenia around today
        """
        death = get_death_or_fallback(self.db, data)
        if death:
            return "%012d" % death.get_date_object().get_sort_value()
        else:
            return "%012d" % 0

    def column_birth_place(self, data):
        event_ref = data.get_birth_ref()
        if event_ref and event_ref.ref:
            event = self.db.get_event_from_handle(event_ref.ref)
            if event:
                return place_displayer.display_event(self.db, event)
        return ""

    def column_death_place(self, data):
        event_ref = data.get_death_ref()
        if event_ref and event_ref.ref:
            event = self.db.get_event_from_handle(event_ref.ref)
            if event:
                return place_displayer.display_event(self.db, event)
        return ""



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Slovak-specific classes for parsing and displaying dates.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import re

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib.date import Date
from ._dateparser import DateParser
from ._datedisplay import DateDisplay
from ._datehandler import register_datehandler


# -------------------------------------------------------------------------
#
# Slovak parser
#
# -------------------------------------------------------------------------
class DateParserSK(DateParser):
    modifier_to_int = {
        "pred": Date.MOD_BEFORE,
        "po": Date.MOD_AFTER,
        "asi": Date.MOD_ABOUT,
        "okolo": Date.MOD_ABOUT,
        "pribl.": Date.MOD_ABOUT,
        "od": Date.MOD_FROM,
        "do": Date.MOD_TO,
    }

    calendar_to_int = {
        "gregoriánsky": Date.CAL_GREGORIAN,
        "g": Date.CAL_GREGORIAN,
        "juliánsky": Date.CAL_JULIAN,
        "j": Date.CAL_JULIAN,
        "hebrejský": Date.CAL_HEBREW,
        "h": Date.CAL_HEBREW,
        "islamský": Date.CAL_ISLAMIC,
        "i": Date.CAL_ISLAMIC,
        "republikánsky": Date.CAL_FRENCH,
        "r": Date.CAL_FRENCH,
        "perzský": Date.CAL_PERSIAN,
        "p": Date.CAL_PERSIAN,
        "švédsky": Date.CAL_SWEDISH,
        "s": Date.CAL_SWEDISH,
    }

    quality_to_int = {
        "odhadovaný": Date.QUAL_ESTIMATED,
        "odh.": Date.QUAL_ESTIMATED,
        "vypočítaný": Date.QUAL_CALCULATED,
        "vyp.": Date.QUAL_CALCULATED,
    }

    def init_strings(self):
        DateParser.init_strings(self)
        _span_1 = ["od"]
        _span_2 = ["do"]
        _range_1 = ["medzi"]
        _range_2 = ["a"]
        self._span = re.compile(
            r"(%s)\s+(?P<start>.+)\s+(%s)\s+(?P<stop>.+)"
            % ("|".join(_span_1), "|".join(_span_2)),
            re.IGNORECASE,
        )
        self._range = re.compile(
            r"(%s)\s+(?P<start>.+)\s+(%s)\s+(?P<stop>.+)"
            % ("|".join(_range_1), "|".join(_range_2)),
            re.IGNORECASE,
        )


# -------------------------------------------------------------------------
#
# Slovak display
#
# -------------------------------------------------------------------------
class DateDisplaySK(DateDisplay):
    """
    Slovak language date display class.
    """

    long_months = (
        "",
        "január",
        "február",
        "marec",
        "apríl",
        "máj",
        "jún",
        "júl",
        "august",
        "september",
        "október",
        "november",
        "december",
    )

    short_months = (
        "",
        "jan",
        "feb",
        "mar",
        "apr",
        "máj",
        "jún",
        "júl",
        "aug",
        "sep",
        "okt",
        "nov",
        "dec",
    )

    calendar = (
        "",
        "juliánsky",
        "hebrejský",
        "republikánsky",
        "perzský",
        "islamský",
        "švédsky",
    )

    _mod_str = ("", "pred ", "po ", "okolo ", "", "", "")

    _qual_str = ("", "odh. ", "vyp. ")

    formats = (
        "RRRR-MM-DD (ISO)",
        "numerický",
        "Mesiac Deň, Rok",
        "MES Deň, Rok",
        "Deň, Mesiac, Rok",
        "Deň MES Rok",
    )
    # this must agree with DateDisplayEn's "formats" definition
    # (since no locale-specific _display_gregorian exists, here)

    def display(self, date):
        """
        Return a text string representing the date.
        """
        mod = date.get_modifier()
        cal = date.get_calendar()
        qual = date.get_quality()
        start = date.get_start_date()
        newyear = date.get_new_year()

        qual_str = self._qual_str[qual]

        if mod == Date.MOD_TEXTONLY:
            return date.get_text()
        elif start == Date.EMPTY:
            return ""
        elif mod == Date.MOD_SPAN:
            d1 = self.display_cal[cal](start)
            d2 = self.display_cal[cal](date.get_stop_date())
            scal = self.format_extras(cal, newyear)
            return "%s%s %s %s %s%s" % (qual_str, "od", d1, "do", d2, scal)
        elif mod == Date.MOD_RANGE:
            d1 = self.display_cal[cal](start)
            d2 = self.display_cal[cal](date.get_stop_date())
            scal = self.format_extras(cal, newyear)
            return "%s%s %s %s %s%s" % (qual_str, "medzi", d1, "a", d2, scal)
        else:
            text = self.display_cal[date.get_calendar()](start)
            scal = self.format_extras(cal, newyear)
            return "%s%s%s%s" % (qual_str, self._mod_str[mod], text, scal)


# -------------------------------------------------------------------------
#
# Register classes
#
# -------------------------------------------------------------------------
register_datehandler(
    ("sk_SK", "sk", "SK", "Slovak", ("%d.%m.%Y",)), DateParserSK, DateDisplaySK
)



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011 Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Gtk modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.lib import EventType, EventRoleType
from gramps.gui.editors import EditEvent
from gramps.gui.listmodel import ListModel, NOSORT
from gramps.gen.plug import Gramplet
from gramps.gen.datehandler import get_date
from gramps.gen.display.place import displayer as place_displayer
from gramps.gen.errors import WindowActiveError
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


class PersonResidence(Gramplet):
    """
    Displays residence events for a person.
    """

    def init(self):
        self.gui.WIDGET = self.build_gui()
        self.gui.get_container_widget().remove(self.gui.textview)
        self.gui.get_container_widget().add(self.gui.WIDGET)
        self.gui.WIDGET.show()

    def build_gui(self):
        """
        Build the GUI interface.
        """
        tip = _("Double-click on a row to edit the selected event.")
        self.set_tooltip(tip)
        top = Gtk.TreeView()
        titles = [
            (
                "",
                NOSORT,
                50,
            ),
            (_("Date"), 1, 200),
            (_("Place"), 2, 200),
        ]
        self.model = ListModel(top, titles, event_func=self.edit_event)
        return top

    def db_changed(self):
        self.connect(self.dbstate.db, "person-update", self.update)

    def active_changed(self, handle):
        self.update()

    def update_has_data(self):
        active_handle = self.get_active("Person")
        if active_handle:
            active = self.dbstate.db.get_person_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def get_has_data(self, active_person):
        """
        Return True if the gramplet has data, else return False.
        """
        if active_person:
            for event_ref in active_person.get_event_ref_list():
                if int(event_ref.get_role()) == EventRoleType.PRIMARY:
                    event = self.dbstate.db.get_event_from_handle(event_ref.ref)
                    if int(event.get_type()) == EventType.RESIDENCE:
                        return True
        return False

    def main(self):  # return false finishes
        self.model.clear()
        active_handle = self.get_active("Person")
        if active_handle:
            active_person = self.dbstate.db.get_person_from_handle(active_handle)
            if active_person:
                self.display_person(active_person)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)

    def display_person(self, active_person):
        """
        Display the residence events of the active person.
        """
        count = 0
        for event_ref in active_person.get_event_ref_list():
            if int(event_ref.get_role()) == EventRoleType.PRIMARY:
                event = self.dbstate.db.get_event_from_handle(event_ref.ref)
                if int(event.get_type()) == EventType.RESIDENCE:
                    self.add_residence(event)
                    count += 1
        self.set_has_data(count > 0)

    def add_residence(self, event):
        """
        Add a residence event to the model.
        """
        date = get_date(event)
        place = ""
        handle = event.get_place_handle()
        if handle:
            place = place_displayer.display_event(self.dbstate.db, event)
        self.model.add((event.get_handle(), date, place))

    def edit_event(self, treeview):
        """
        Edit the selected event.
        """
        model, iter_ = treeview.get_selection().get_selected()
        if iter_:
            handle = model.get_value(iter_, 0)
            try:
                event = self.dbstate.db.get_event_from_handle(handle)
                EditEvent(self.dbstate, self.uistate, [], event)
            except WindowActiveError:
                pass



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2017       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Serialization utilities for Gramps.
"""

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------
import json

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
import gramps.gen.lib as lib


def __default(obj):
    obj_dict = {"_class": obj.__class__.__name__}
    if isinstance(obj, lib.GrampsType):
        obj_dict["string"] = getattr(obj, "string")
    if isinstance(obj, lib.Date):
        if obj.is_empty() and not obj.text:
            return None
    for key, value in obj.__dict__.items():
        if not key.startswith("_"):
            obj_dict[key] = value
    for key, value in obj.__class__.__dict__.items():
        if isinstance(value, property):
            if key != "year":
                obj_dict[key] = getattr(obj, key)
    return obj_dict


def __object_hook(obj_dict):
    obj = getattr(lib, obj_dict["_class"])()
    for key, value in obj_dict.items():
        if key != "_class":
            if key in ("dateval", "rect") and value is not None:
                value = tuple(value)
            if key == "ranges":
                value = [tuple(item) for item in value]
            setattr(obj, key, value)
    if obj_dict["_class"] == "Date":
        if obj.is_empty() and not obj.text:
            return None
    return obj


def to_json(obj):
    """
    Encode a Gramps object in JSON format.

    :param obj: The object to be serialized.
    :type obj: object
    :returns: A JSON string.
    :rtype: str
    """
    return json.dumps(obj, default=__default, ensure_ascii=False)


def from_json(data):
    """
    Decode JSON data into a Gramps object hierarchy.

    :param data: The JSON string to be unserialized.
    :type data: str
    :returns: A Gramps object.
    :rtype: object
    """
    return json.loads(data, object_hook=__object_hook)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext

# -------------------------------------------------------------------------
#
# GNOME modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import GObject

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.plug.docgen import PaperStyle, PAPER_PORTRAIT, PAPER_LANDSCAPE
from gramps.gen.plug.report._paper import paper_sizes
from ...glade import Glade


# -------------------------------------------------------------------------
#
# PaperComboBox
#
# -------------------------------------------------------------------------
class PaperComboBox(Gtk.ComboBox):
    def __init__(self, default_name):
        Gtk.ComboBox.__init__(self)
        self.set_wrap_width(4)

        self.store = Gtk.ListStore(GObject.TYPE_STRING)
        self.set_model(self.store)
        cell = Gtk.CellRendererText()
        self.pack_start(cell, True)
        self.add_attribute(cell, "text", 0)
        self.mapping = {}

        index = 0
        start_index = 0
        for key in paper_sizes:
            key_name = key.get_name()
            if default_name == key_name or default_name == key.trans_pname:
                start_index = index
            self.mapping[key_name] = key  # always use the English paper name
            if key.trans_pname:
                key_name = key.trans_pname  # display the translated paper name
            self.store.append(row=[key_name])
            index += 1

        self.set_active(start_index)

    def get_value(self):
        active = self.get_active()
        if active < 0:
            return None
        key = str(self.store[active][0])
        for paper in paper_sizes:
            if key == paper.trans_pname:
                key = paper.get_name()
        return (self.mapping[key], key)


# -------------------------------------------------------------------------
#
# OrientationComboBox
#
# -------------------------------------------------------------------------
class OrientationComboBox(Gtk.ComboBox):
    def __init__(self, default=PAPER_PORTRAIT):
        Gtk.ComboBox.__init__(self)

        self.store = Gtk.ListStore(GObject.TYPE_STRING)
        self.set_model(self.store)
        cell = Gtk.CellRendererText()
        self.pack_start(cell, True)
        self.add_attribute(cell, "text", 0)
        self.mapping = {}

        self.store.append(row=[_("Portrait")])
        self.store.append(row=[_("Landscape")])
        if default == PAPER_PORTRAIT:
            self.set_active(0)
        else:
            self.set_active(1)

    def set_value(self, value=0):
        if value == PAPER_PORTRAIT:
            self.set_active(0)
        else:
            self.set_active(1)

    def get_value(self):
        active = self.get_active()
        if active < 0:
            return None
        if active == 0:
            return PAPER_PORTRAIT
        else:
            return PAPER_LANDSCAPE


# -------------------------------------------------------------------------
#
# PaperFrame
#
# -------------------------------------------------------------------------
class PaperFrame(Gtk.Box):
    """PaperFrame provides all the entry necessary to specify a paper style."""

    def __init__(
        self,
        default_metric,
        default_name,
        default_orientation,
        margins=[2.54, 2.54, 2.54, 2.54],
        custom=[29.7, 21.0],
    ):
        Gtk.Box.__init__(self)
        glade_xml = Glade()

        self.paper_grid = glade_xml.get_object("paper_grid")
        self.top = glade_xml.get_object("papermenu")

        # get all the widgets
        widgets = (
            "pwidth",
            "pheight",
            "lmargin",
            "rmargin",
            "tmargin",
            "bmargin",
            "lunits1",
            "lunits2",
            "lunits3",
            "lunits4",
            "lunits5",
            "lunits6",
            "metric",
        )

        for w in widgets:
            setattr(self, w, glade_xml.get_object(w))

        # insert custom widgets
        self.papersize_menu = PaperComboBox(default_name)
        self.orientation_menu = OrientationComboBox(default_orientation)
        self.metric.set_active(default_metric)

        # connect all widgets
        format_grid = glade_xml.get_object("format_grid")
        format_grid.attach(self.papersize_menu, 1, 0, 1, 1)
        format_grid.attach(self.orientation_menu, 1, 3, 1, 1)

        # connect signals
        self.papersize_menu.connect("changed", self.size_changed)
        self.metric.connect("toggled", self.units_changed)

        # set initial values
        self.paper_unit = "cm"
        self.paper_unit_multiplier = 1.0

        self.pwidth.set_text("%.2f" % custom[0])
        self.pheight.set_text("%.2f" % custom[1])
        self.lmargin.set_text("%.2f" % margins[0])
        self.rmargin.set_text("%.2f" % margins[1])
        self.tmargin.set_text("%.2f" % margins[2])
        self.bmargin.set_text("%.2f" % margins[3])

        self.paper_grid.show_all()
        # Shift the grid from glade toplevel window to this box
        self.paper_grid.get_parent().remove(self.paper_grid)
        self.add(self.paper_grid)
        # need to get rid of glade toplevel now that we are done with it.
        self.top.destroy()

        self.units_changed(self.metric)
        self.size_changed(None)

    def size_changed(self, obj):
        """Paper size combobox 'changed' callback."""
        size, name = self.get_paper_size()

        is_custom = name == "Custom Size"
        self.pwidth.set_sensitive(is_custom)
        self.pheight.set_sensitive(is_custom)

        if self.paper_unit == "cm":
            self.pwidth.set_text("%.2f" % size.get_width())
            self.pheight.set_text("%.2f" % size.get_height())
        elif self.paper_unit == "in.":
            self.pwidth.set_text("%.2f" % size.get_width_inches())
            self.pheight.set_text("%.2f" % size.get_height_inches())
        else:
            raise ValueError(
                'Paper dimension unit "%s" is not allowed' % self.paper_unit
            )

    def units_changed(self, checkbox):
        """Metric checkbox 'toggled' callback."""
        paper_size, paper_name = self.get_paper_size()
        paper_margins = self.get_paper_margins()

        if checkbox.get_active():
            self.paper_unit = "cm"
            self.paper_unit_multiplier = 1.0
            paper_unit_text = _("cm")
        else:
            self.paper_unit = "in."
            self.paper_unit_multiplier = 2.54
            paper_unit_text = _("in.", "inch")

        self.lunits1.set_text(paper_unit_text)
        self.lunits2.set_text(paper_unit_text)
        self.lunits3.set_text(paper_unit_text)
        self.lunits4.set_text(paper_unit_text)
        self.lunits5.set_text(paper_unit_text)
        self.lunits6.set_text(paper_unit_text)

        if self.paper_unit == "cm":
            self.pwidth.set_text("%.2f" % paper_size.get_width())
            self.pheight.set_text("%.2f" % paper_size.get_height())
        else:
            self.pwidth.set_text("%.2f" % paper_size.get_width_inches())
            self.pheight.set_text("%.2f" % paper_size.get_height_inches())

        self.lmargin.set_text("%.2f" % (paper_margins[0] / self.paper_unit_multiplier))
        self.rmargin.set_text("%.2f" % (paper_margins[1] / self.paper_unit_multiplier))
        self.tmargin.set_text("%.2f" % (paper_margins[2] / self.paper_unit_multiplier))
        self.bmargin.set_text("%.2f" % (paper_margins[3] / self.paper_unit_multiplier))

    def get_paper_size(self):
        """Read and validate paper size values.

        If needed update the dimensions from the width, height entries,
        and worst case fallback to A4 size.

        """
        papersize, papername = self.papersize_menu.get_value()
        if papername == "Custom Size":
            try:
                h = float(str(self.pheight.get_text().replace(",", ".")))
                w = float(str(self.pwidth.get_text().replace(",", ".")))

                if h <= 1.0 or w <= 1.0:
                    papersize.set_height(29.7)
                    papersize.set_width(21.0)
                else:
                    papersize.set_height(h * self.paper_unit_multiplier)
                    papersize.set_width(w * self.paper_unit_multiplier)
            except:
                papersize.set_height(29.7)
                papersize.set_width(21.0)

        return papersize, papername

    def get_paper_margins(self):
        """Get and validate margin values from dialog entries.

        Values returned in [cm].

        """
        paper_margins = [
            str(margin.get_text())
            for margin in (self.lmargin, self.rmargin, self.tmargin, self.bmargin)
        ]

        for i, margin in enumerate(paper_margins):
            try:
                paper_margins[i] = float(margin.replace(",", "."))
                paper_margins[i] = paper_margins[i] * self.paper_unit_multiplier
                paper_margins[i] = max(paper_margins[i], 0)
            except:
                paper_margins[i] = 2.54

        return paper_margins

    def get_custom_paper_size(self):
        """Get and validate custom paper size values from dialog entries.
        Float values returned.
        """
        try:
            width = (
                float(self.pwidth.get_text().replace(",", "."))
                * self.paper_unit_multiplier
            )
            height = (
                float(self.pheight.get_text().replace(",", "."))
                * self.paper_unit_multiplier
            )
        except ValueError:
            width = float(21.0)
            height = float(29.7)

        paper_size = [max(width, 1.0), max(height, 1.0)]

        return paper_size

    def get_paper_style(self):
        paper_size, paper_name = self.get_paper_size()
        paper_orientation = self.orientation_menu.get_value()
        paper_margins = self.get_paper_margins()

        pstyle = PaperStyle(paper_size, paper_orientation, *paper_margins)
        return pstyle

    def get_paper_metric(self):
        return self.metric.get_active()

    def get_paper_name(self):
        paper_size, paper_name = self.get_paper_size()
        return paper_name

    def get_orientation(self):
        return self.orientation_menu.get_value()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009 Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Html and Html format management for the different reports
"""

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------
from xml.sax.saxutils import escape
import os.path

# -------------------------------------------------------------------------
#
# GTK modules
#
# -------------------------------------------------------------------------

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.plug.docbackend import DocBackend
from gramps.plugins.lib.libhtml import Html, xml_lang
from gramps.gen.errors import ReportError

from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# ------------------------------------------------------------------------
#
# Functions
#
# ------------------------------------------------------------------------


def process_spaces(intext, format):
    """
    Function to process spaces in text lines for pre-formatted notes.
    line : text to process
    format : = 0 : Flowed, = 1 : Preformatted

    If the text is pre-formatted (format==1), then leading spaces  (after ignoring XML)
    are replaced by alternating non-breaking spaces and ordinary spaces.
    After the first non-space character, single spaces are left
    but multiple spaces are replaced by alternating NBSP and space
    If the text is flowed, the text is unchanged.

    Returns the processed text, and the number of significant
    (i.e. non-xml non-white-space) chars.
    """
    NORMAL = 1
    SPACE = 2
    NBSP = 3
    XML = 4
    SPACEHOLD = 5

    sigcount = 0
    state = NORMAL
    outtext = ""
    if format == 1:
        # Pre-formatted
        for char in intext:
            if state == NORMAL:
                if char == " ":
                    if sigcount == 0:
                        state = NBSP
                        outtext += "&nbsp;"
                    else:
                        state = SPACEHOLD
                elif char == "<":
                    state = XML
                    outtext += char
                else:
                    sigcount += 1
                    outtext += char
            elif state == SPACE:
                if char == " ":
                    state = NBSP
                    outtext += "&nbsp;"
                elif char == "<":
                    state = XML
                    outtext += char
                else:
                    sigcount += 1
                    state = NORMAL
                    outtext += char
            elif state == NBSP:
                if char == " ":
                    state = SPACE
                elif char == "<":
                    state = XML
                else:
                    sigcount += 1
                    state = NORMAL
                outtext += char
            elif state == XML:
                if char == ">":
                    state = NORMAL
                outtext += char
            elif state == SPACEHOLD:
                if char == " ":
                    outtext += " &nbsp;"
                    state = NORMAL
                elif char == "<":
                    outtext += " " + char
                    state = XML
                else:
                    outtext += " " + char
                    sigcount += 1
                    state = NORMAL

    else:
        # format == 0 flowed
        for char in intext:
            if char == "<" and state == NORMAL:
                state = XML
                outtext += char
            elif char == ">" and state == XML:
                state = NORMAL
                outtext += char
            elif state == XML:
                outtext += char
            else:
                sigcount += 1
                outtext += char

    return [outtext, sigcount]


# ------------------------------------------------------------------------
#
# Document Backend class for html pages
#
# ------------------------------------------------------------------------


class HtmlBackend(DocBackend):
    """
    Implementation for html pages
    Contrary to other backends, we do not write to file but to a Html object
    instead, writing out the file on close.
    """

    STYLETAG_TO_PROPERTY = {
        DocBackend.FONTCOLOR: "color:%s;",
        DocBackend.HIGHLIGHT: "background-color:%s;",
        DocBackend.FONTFACE: "font-family:'%s';",
        DocBackend.FONTSIZE: "font-size:%spx;",
    }

    # overwrite base class attributes, they become static var of CairoDoc
    SUPPORTED_MARKUP = [
        DocBackend.BOLD,
        DocBackend.ITALIC,
        DocBackend.UNDERLINE,
        DocBackend.FONTFACE,
        DocBackend.FONTSIZE,
        DocBackend.FONTCOLOR,
        DocBackend.HIGHLIGHT,
        DocBackend.SUPERSCRIPT,
        DocBackend.LINK,
        DocBackend.STRIKETHROUGH,
        DocBackend.SUBSCRIPT,
    ]

    STYLETAG_MARKUP = {
        DocBackend.BOLD: ("<strong>", "</strong>"),
        DocBackend.ITALIC: ("<em>", "</em>"),
        DocBackend.UNDERLINE: ('<span style="text-decoration:underline;">', "</span>"),
        DocBackend.SUPERSCRIPT: ("<sup>", "</sup>"),
        DocBackend.STRIKETHROUGH: ("<s>", "</s>"),
        DocBackend.SUBSCRIPT: ("<sub>", "</sub>"),
    }

    ESCAPE_FUNC = lambda self: escape

    def __init__(self, filename=None):
        """
        @param filename: path name of the file the backend works on
        """
        DocBackend.__init__(self, filename)
        self.html_page = None
        self.html_header = None
        self.html_body = None
        self._subdir = None
        self.title = None
        self.build_link = None

    def _create_xmltag(self, tagtype, value):
        """
        overwrites the method in DocBackend
        creates the pango xml tags needed for non bool style types
        """
        if tagtype not in self.SUPPORTED_MARKUP:
            return None
        elif tagtype == DocBackend.LINK:
            return self.format_link(value)
        elif tagtype == DocBackend.FONTSIZE:
            # size is in points
            value = str(value)
        elif tagtype == DocBackend.FONTFACE:
            # fonts can have strange symbols in them, ' needs to be escaped
            value = value.replace("'", "\\'") if value else ""
        return (
            '<span style="%s">' % (self.STYLETAG_TO_PROPERTY[tagtype] % (value)),
            "</span>",
        )

    def _checkfilename(self):
        """
        Check to make sure filename satisfies the standards for this filetype
        """
        fparts = os.path.basename(self._filename).split(".")
        if not len(fparts) >= 2 and not (
            fparts[-1] == "html" or fparts[-1] == "htm" or fparts[-1] == "php"
        ):
            self._filename = self._filename + ".htm"
        fparts = os.path.basename(self._filename).split(".")
        self._subdir = ".".join(fparts[:-1])

    def set_title(self, title):
        """
        Set the title to use for the html page
        """
        self.title = title

    def open(self):
        """
        overwrite method, htmlbackend creates a html object that is written on
        close
        """
        try:
            DocBackend.open(self)
        except IOError as msg:
            errmsg = "%s\n%s" % (_("Could not create %s") % self._filename, msg)
            raise ReportError(errmsg)
        except:
            raise ReportError(_("Could not create %s") % self._filename)
        if not os.path.isdir(self.datadirfull()):
            try:
                os.mkdir(self.datadirfull())
            except IOError as msg:
                errmsg = "%s\n%s" % (_("Could not create %s") % self.datadirfull(), msg)
                raise ReportError(errmsg)
            except:
                raise ReportError(_("Could not create %s") % self.datadirfull())
        self.html_page, self.html_header, self.html_body = Html.page(
            lang=xml_lang(), title=self.title
        )

    def __write(self, string):
        """a write to the file"""
        DocBackend.write(self, string + "\n")

    def write(self, obj):
        """write to the html page. One can pass a html object, or a string"""
        self.html_body += obj

    def close(self):
        """
        write out the html to the page
        """
        self.html_page.write(self.__write, indent="  ")
        DocBackend.close(self)

    def datadir(self):
        """
        the directory where to save extra files
        """
        return self._subdir

    def datadirfull(self):
        """
        full path of the datadir directory
        """
        return os.path.join(os.path.dirname(self.getf()), self.datadir())

    def format_link(self, value):
        """
        Override of base method.
        """
        if value.startswith("gramps://"):
            if self.build_link:
                obj_class, prop, handle = value[9:].split("/", 2)
                if prop in ["handle", "gramps_id"]:
                    value = self.build_link(prop, handle, obj_class)
                    if not value:
                        return self.STYLETAG_MARKUP[DocBackend.UNDERLINE]
                else:
                    return self.STYLETAG_MARKUP[DocBackend.UNDERLINE]
            else:
                return self.STYLETAG_MARKUP[DocBackend.UNDERLINE]
        return ('<a href="%s">' % self.ESCAPE_FUNC()(value), "</a>")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2010       Nick Hall
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
DbBookmarks Class
"""

from ..utils.callback import Callback


# -------------------------------------------------------------------------
#
# DbBookmarks class
#
# -------------------------------------------------------------------------
class DbBookmarks(Callback):
    """
    Simple class for tracking a list of bookmarked objects
    """

    __signals__ = {"bookmark-list-changed": None}

    def __init__(self, default=None):
        super().__init__()
        self.bookmarks = default or []

    def load(self, new_list):
        """
        Quietly load the bookmark list.
        """
        self.bookmarks = list(new_list)

    def set(self, new_list):
        """
        Set the bookmark list.
        """
        self.bookmarks = list(new_list)
        self.emit("bookmark-list-changed")

    def get(self):
        """
        Return the bookmark list.
        """
        return self.bookmarks

    def append(self, item):
        """
        Append an item to the end of the list.
        """
        self.bookmarks.append(item)
        self.emit("bookmark-list-changed")

    def append_list(self, blist):
        """
        Append a list of items to the end of the list.
        """
        self.bookmarks += blist
        self.emit("bookmark-list-changed")

    def remove(self, item):
        """
        Remove an item from the list.
        """
        self.bookmarks.remove(item)
        self.emit("bookmark-list-changed")

    def pop(self, index):
        """
        Pop an item from the list.
        """
        item = self.bookmarks.pop(index)
        self.emit("bookmark-list-changed")
        return item

    def insert(self, pos, item):
        """
        Insert a new item into the list.
        """
        self.bookmarks.insert(pos, item)
        self.emit("bookmark-list-changed")

    def close(self):
        """
        Close the list by deleting it.
        """
        del self.bookmarks



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2019  Matthias Kemmer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ...lib.attrtype import AttributeType
from . import Rule


# -------------------------------------------------------------------------
#
# HasAttribute
#
# -------------------------------------------------------------------------
class HasAttributeBase(Rule):
    """
    Rule that checks for an object with a particular attribute.
    """

    labels = ["Attribute:", "Value:"]
    name = "Objects with the <attribute>"
    description = "Matches objects with the given attribute " "of a particular value"
    category = _("General filters")
    allow_regex = True

    def apply(self, db, obj):
        if not self.list[0]:
            return False
        for attr in obj.get_attribute_list():
            specified_type = AttributeType()
            specified_type.set_from_xml_str(self.list[0])
            name_match = attr.get_type() == specified_type

            if name_match:
                if self.match_substring(1, attr.get_value()):
                    return True
        return False
